#!/usr/bin/env python
# -*- coding: utf-8 -*-

from PyQt4 import QtCore, QtGui, QtOpenGL


class Widget(QtOpenGL.QGLWidget):
    def __init__(self, parent, controller):
        QtOpenGL.QGLWidget.__init__(
                self, 
                QtOpenGL.QGLFormat(QtOpenGL.QGL.SampleBuffers), 
                parent)
        self.controller=controller

    def minimumSizeHint(self):
        return QtCore.QSize(50, 50)

    def sizeHint(self):
        return QtCore.QSize(150, 150)

    def paintGL(self):
        self.controller.draw()

    def resizeGL(self, width, height):
        self.controller.onResize(width, height)

    def mousePressEvent(self, event):
        if event.button()==QtCore.Qt.LeftButton:
            if self.controller.onLeftDown(event.x(), event.y()):
                self.repaint()
        elif event.button()==QtCore.Qt.MiddleButton:
            if self.controller.onMiddleDown(event.x(), event.y()):
                self.repaint()
        elif event.button()==QtCore.Qt.RightButton:
            if self.controller.onRightDown(event.x(), event.y()):
                self.repaint()

    def mouseReleaseEvent(self, event):
        if event.button()==QtCore.Qt.LeftButton:
            if self.controller.onLeftUp(event.x(), event.y()):
                self.repaint()
        elif event.button()==QtCore.Qt.MiddleButton:
            if self.controller.onMiddleUp(event.x(), event.y()):
                self.repaint()
        elif event.button()==QtCore.Qt.RightButton:
            if self.controller.onRightUp(event.x(), event.y()):
                self.repaint()

    def mouseMoveEvent(self, event):
        if self.controller.onMotion(event.x(), event.y()):
            self.repaint()

    def wheelEvent(self, event):
        if self.controller.onWheel(-event.delta()):
            self.repaint()


if __name__=="__main__":
    from PyQt4 import Qt
    import glglue.sample
    class Window(Qt.QWidget):
        def __init__(self, parent=None):
            Qt.QWidget.__init__(self, parent)
            # setup opengl widget
            self.controller=glglue.sample.SampleController()
            self.glwidget=Widget(self, self.controller)
            # packing
            mainLayout = Qt.QHBoxLayout()
            mainLayout.addWidget(self.glwidget)
            self.setLayout(mainLayout)

    import sys
    app = Qt.QApplication(sys.argv)
    window = Window()
    window.show()
    sys.exit(app.exec_())




from typing import Optional
import sys
from logging import getLogger
from OpenGL import GL
import OpenGL.error
from ctypes import wintypes
import ctypes
from glglue import ctypesmath

from glglue.basecontroller import BaseController
from . import win32con
logger = getLogger(__name__)


##############################################################################
# Windows types
##############################################################################
def ErrorIfZero(handle):
    if handle == 0:
        raise ctypes.WinError()
    else:
        return handle


def LOWORD(n):
    return n & 0xFFFF


def HIWORD(n):
    return (n >> 16) & 0XFFFF


WNDPROC = ctypes.WINFUNCTYPE(
    ctypes.c_int, wintypes.HWND, ctypes.c_uint, wintypes.WPARAM, wintypes.LPARAM)


class WNDCLASS(ctypes.Structure):
    _fields_ = [('style', ctypes.c_uint), ('lpfnWndProc', WNDPROC),
                ('cbClsExtra', ctypes.c_int), ('cbWndExtra', ctypes.c_int),
                ('hInstance', wintypes.HANDLE), ('hIcon',
                                                 wintypes.HANDLE), ('hCursor', wintypes.HANDLE),
                ('hbrBackground', wintypes.HANDLE), ('lpszMenuName', wintypes.LPCSTR),
                ('lpszClassName', wintypes.LPCSTR)]


class CREATESTRUCT(ctypes.Structure):
    _fields_ = [('lpCreateParams', ctypes.c_void_p), ('hInstance', wintypes.HANDLE),
                ('hMenu', wintypes.HANDLE), ('hwndParent',
                                             wintypes.HWND), ('cy', ctypes.c_int),
                ('cx', ctypes.c_int), ('y', ctypes.c_int), ('x',
                                                            ctypes.c_int), ('style', ctypes.c_int),
                ('lpszName', wintypes.LPCSTR), ('lpszClass', wintypes.LPCSTR),
                ('dwExStyle', ctypes.c_int)]


class PIXELFORMATDESCRIPTOR(ctypes.Structure):
    _fields_ = [
        ('nSize', ctypes.c_ushort),
        ('nVersion', ctypes.c_ushort),
        ('dwFlags', ctypes.c_ulong),
        ('iPixelType', ctypes.c_ubyte),
        ('cColorBits', ctypes.c_ubyte),
        ('cRedBits', ctypes.c_ubyte),
        ('cRedShift', ctypes.c_ubyte),
        ('cGreenBits', ctypes.c_ubyte),
        ('cGreenShift', ctypes.c_ubyte),
        ('cBlueBits', ctypes.c_ubyte),
        ('cBlueShift', ctypes.c_ubyte),
        ('cAlphaBits', ctypes.c_ubyte),
        ('cAlphaShift', ctypes.c_ubyte),
        ('cAccumBits', ctypes.c_ubyte),
        ('cAccumRedBits', ctypes.c_ubyte),
        ('cAccumGreenBits', ctypes.c_ubyte),
        ('cAccumBlueBits', ctypes.c_ubyte),
        ('cAccumAlphaBits', ctypes.c_ubyte),
        ('cDepthBits', ctypes.c_ubyte),
        ('cStencilBits', ctypes.c_ubyte),
        ('cAuxBuffers', ctypes.c_ubyte),
        ('iLayerType', ctypes.c_ubyte),
        ('bReserved', ctypes.c_ubyte),
        ('dwLayerMask', ctypes.c_ulong),
        ('dwVisibleMask', ctypes.c_ulong),
        ('dwDamageMask', ctypes.c_ulong),
    ]


class RECT(ctypes.Structure):
    _fields_ = [('left', ctypes.c_long), ('top', ctypes.c_long), ('right', ctypes.c_long),
                ('bottom', ctypes.c_long)]


class PAINTSTRUCT(ctypes.Structure):
    _fields_ = [('hdc', wintypes.HDC), ('fErase', wintypes.BOOL), ('rcPaint', RECT),
                ('fRestore', wintypes.BOOL), ('fIncUpdate', wintypes.BOOL),
                ('rgbReserved', ctypes.c_char * 32)]


class POINT(ctypes.Structure):
    _fields_ = [('x', ctypes.c_long), ('y', ctypes.c_long)]


class MSG(ctypes.Structure):
    _fields_ = [('hwnd', wintypes.HWND), ('message', ctypes.c_uint), ('wParam', wintypes.WPARAM),
                ('lParam', wintypes.LPARAM), ('time', ctypes.c_int), ('pt', POINT)]


ShowWindow = ctypes.windll.user32.ShowWindow
ShowWindow.argtypes = [wintypes.HWND, ctypes.c_int]
ShowWindow.restype = wintypes.BOOL

RegisterClass = ctypes.windll.user32.RegisterClassA
RegisterClass.argtypes = [ctypes.c_void_p]
RegisterClass.restype = wintypes.ATOM

CreateWindowEx = ctypes.windll.user32.CreateWindowExA
CreateWindowEx.argtypes = [
    ctypes.c_int, wintypes.LPCSTR, wintypes.LPCSTR, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int, wintypes.HWND, wintypes.HANDLE,
    wintypes.HANDLE, ctypes.py_object
]
CreateWindowEx.restype = wintypes.HWND

if sys.maxsize > 2**32:
    # 64bit
    SetWindowLongPtr = ctypes.windll.user32.SetWindowLongPtrA
    SetWindowLongPtr.argtypes = [wintypes.HWND,
                                 ctypes.c_int, ctypes.c_void_p]
    SetWindowLongPtr.restype = ctypes.c_void_p

    GetWindowLongPtr = ctypes.windll.user32.GetWindowLongPtrA
    GetWindowLongPtr.argtypes = [wintypes.HWND, ctypes.c_int]
    GetWindowLongPtr.restype = ctypes.c_void_p
else:
    # 64bit
    SetWindowLongPtr = ctypes.windll.user32.SetWindowLongA
    SetWindowLongPtr.argtypes = [wintypes.HWND,
                                 ctypes.c_int, ctypes.c_void_p]
    SetWindowLongPtr.restype = ctypes.c_void_p

    GetWindowLongPtr = ctypes.windll.user32.GetWindowLongA
    GetWindowLongPtr.argtypes = [wintypes.HWND, ctypes.c_int]
    GetWindowLongPtr.restype = ctypes.c_void_p

GetDC = ctypes.windll.user32.GetDC
GetDC.argtypes = [wintypes.HWND]

BeginPaint = ctypes.windll.user32.BeginPaint
BeginPaint.argtypes = [wintypes.HWND, ctypes.c_void_p]
BeginPaint.restype = wintypes.HDC

EndPaint = ctypes.windll.user32.EndPaint
EndPaint.argtypes = [wintypes.HWND, ctypes.c_void_p]
EndPaint.restype = wintypes.HDC

ChoosePixelFormat = ctypes.windll.gdi32.ChoosePixelFormat
ChoosePixelFormat.argtypes = [wintypes.HDC, ctypes.c_void_p]
ChoosePixelFormat.restype = ErrorIfZero

SetPixelFormat = ctypes.windll.gdi32.SetPixelFormat
SetPixelFormat.argtypes = [
    wintypes.HDC,
    ctypes.c_int,
    ctypes.c_void_p,
]
SetPixelFormat.restype = ErrorIfZero

SwapBuffers = ctypes.windll.gdi32.SwapBuffers
SwapBuffers.argtypes = [
    wintypes.HDC,
]
SwapBuffers.restype = ErrorIfZero

wglCreateContext = ctypes.windll.opengl32.wglCreateContext
wglCreateContext.argtypes = [
    wintypes.HDC,
]
wglCreateContext.restype = ErrorIfZero

wglDeleteContext = ctypes.windll.opengl32.wglDeleteContext

wglMakeCurrent = ctypes.windll.opengl32.wglMakeCurrent
wglMakeCurrent.argtypes = [
    wintypes.HDC,
    wintypes.HANDLE,
]
wglMakeCurrent.restype = ErrorIfZero

wglGetProcAddress = ctypes.windll.opengl32.wglGetProcAddress
wglGetProcAddress.argtypes = [ctypes.c_void_p]
wglGetProcAddress.restype = ErrorIfZero

DefWindowProc = ctypes.windll.user32.DefWindowProcA
DefWindowProc.argtypes = [wintypes.HWND,
                          ctypes.c_int, wintypes.WPARAM, wintypes.LPARAM]
DefWindowProc.restype = ctypes.c_int

timeGetTime = ctypes.windll.winmm.timeGetTime
timeGetTime.argtypes = []
timeGetTime.restype = ctypes.c_ulong

##############################################################################
# api class
##############################################################################


class Window(object):
    def __init__(self):
        self.hwnd = None
        self.hrc = None
        self.controller: Optional[BaseController] = None
        self.ptrs = {}
        self.callbacks = {
            win32con.WM_PAINT: self.onPaint,
            win32con.WM_SIZE: self.onSize,
            win32con.WM_ERASEBKGND: self.onErasebkgnd,
            win32con.WM_DESTROY: self.onDestroy,
            win32con.WM_LBUTTONDOWN: self.onLeftDown,
            win32con.WM_LBUTTONUP: self.onLeftUp,
            win32con.WM_MBUTTONDOWN: self.onMiddleDown,
            win32con.WM_MBUTTONUP: self.onMiddleUp,
            win32con.WM_RBUTTONDOWN: self.onRightDown,
            win32con.WM_RBUTTONUP: self.onRightUp,
            win32con.WM_MOUSEMOVE: self.onMouseMove,
            win32con.WM_MOUSEWHEEL: self.onMouseWheel,
            win32con.WM_KEYDOWN: self.onKeyDown,
        }
        self.mouse_left = False
        self.mouse_right = False
        self.mouse_middle = False
        self.mouse_capture = False

    def finalize(self):
        logger.debug('finalize %s', self.__class__)

    def update_mouse(self, left, middle, right):
        self.mouse_left = left
        self.mouse_middle = middle
        self.mouse_right = right
        if left or middle or right:
            if not self.mouse_capture:
                ctypes.windll.user32.SetCapture(self.hwnd)
                self.mouse_capture = True
        else:
            if self.mouse_capture:
                ctypes.windll.user32.ReleaseCapture()
                self.mouse_capture = False

    def Redraw(self):
        ctypes.windll.user32.InvalidateRect(
            self.hwnd, ctypes.c_int(0), ctypes.c_int(0))

    def Show(self):
        ShowWindow(self.hwnd, win32con.SW_SHOWNORMAL)
        ctypes.windll.user32.UpdateWindow(self.hwnd)

    def SetLongPtr(self, hwnd, key, value):
        self.ptrs[key] = value
        SetWindowLongPtr(hwnd, key, value)

    def onPaint(self, hwnd, message, wParam, lParam):
        ps = PAINTSTRUCT()
        hdc = BeginPaint(hwnd, ctypes.byref(ps))
        # if self.hrc and self.controller:
        #     self.controller.draw()
        #     SwapBuffers(hdc)
        EndPaint(hwnd, ctypes.byref(ps))
        return 0

    def onSize(self, hwnd, message, wParam, lParam):
        w = LOWORD(lParam)
        h = HIWORD(lParam)
        # logger.debug("WM_SIZE %d x %d %s" % (w, h, self.controller))
        if self.controller:
            self.controller.onResize(w, h)
        return 0

    def onErasebkgnd(self, hwnd, message, wParam, lParam):
        return 0

    def onDestroy(self, hwnd, message, wParam, lParam):
        # sys.stderr.write("WM_DESTROY\n")
        ctypes.windll.user32.PostQuitMessage(0)
        return 0

    def onKeyDown(self, hwnd, message, wParam, lParam):
        if self.controller:
            if wParam >= 65 and wParam <= 90:
                # lower
                if self.controller.onKeyDown(wParam + 32):
                    self.Redraw()
            else:
                if self.controller.onKeyDown(wParam):
                    self.Redraw()
        return 0

    def onMouseMove(self, hwnd, message, wParam, lParam):
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onMotion(x, y):
                self.Redraw()
        return 0

    def onMouseWheel(self, hwnd, message, wParam, lParam):
        if self.controller:
            d = HIWORD(wParam)
            if self.controller.onWheel(d > 32767 and 65536 - d or -d):
                self.Redraw()
        return 0

    def onLeftUp(self, hwnd, message, wParam, lParam):
        self.update_mouse(False, self.mouse_middle, self.mouse_right)
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onLeftUp(x, y):
                self.Redraw()
        return 0

    def onLeftDown(self, hwnd, message, wParam, lParam):
        self.update_mouse(True, self.mouse_middle, self.mouse_right)
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onLeftDown(x, y):
                self.Redraw()
        return 0

    def onMiddleUp(self, hwnd, message, wParam, lParam):
        self.update_mouse(self.mouse_left, False, self.mouse_right)
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onMiddleUp(x, y):
                self.Redraw()
        return 0

    def onMiddleDown(self, hwnd, message, wParam, lParam):
        self.update_mouse(self.mouse_left, True, self.mouse_right)
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onMiddleDown(x, y):
                self.Redraw()
        return 0

    def onRightUp(self, hwnd, message, wParam, lParam):
        self.update_mouse(self.mouse_left, self.mouse_middle, False)
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onRightUp(x, y):
                self.Redraw()
        return 0

    def onRightDown(self, hwnd, message, wParam, lParam):
        self.update_mouse(self.mouse_left, self.mouse_middle, True)
        if self.controller:
            x = LOWORD(lParam)
            y = HIWORD(lParam)
            if self.controller.onRightDown(x, y):
                self.Redraw()
        return 0

    def WndProc(self, hwnd, message, wParam, lParam):
        if message in self.callbacks:
            return self.callbacks[message](hwnd, message, wParam, lParam)
        else:
            return DefWindowProc(hwnd, message, wParam, lParam)

    def createGLContext(self, bits, version=None, profile='compatibility'):
        hdc = GetDC(self.hwnd)

        pfd = PIXELFORMATDESCRIPTOR()
        pfd.nsize = ctypes.sizeof(PIXELFORMATDESCRIPTOR)
        pfd.nVersion = 1
        pfd.dwFlags = (win32con.PFD_DRAW_TO_WINDOW
                       | win32con.PFD_SUPPORT_OPENGL
                       | win32con.PFD_DOUBLEBUFFER)
        pfd.iPixelType = win32con.PFD_TYPE_RGBA
        pfd.cColorBits = bits
        pfd.cRedBits = 0
        pfd.cRedShift = 0
        pfd.cGreenBits = 0
        pfd.cGreenShift = 0
        pfd.cBlueBits = 0
        pfd.cBlueShift = 0
        pfd.cAlphaBits = 0
        pfd.cAlphaShift = 0
        pfd.cAccumBits = 0
        pfd.cAccumRedBits = 0
        pfd.cAccumGreenBits = 0
        pfd.cAccumBlueBits = 0
        pfd.cAccumAlphaBits = 0
        pfd.cDepthBits = bits
        pfd.cStencilBits = 2
        pfd.cAuxBuffers = 0
        pfd.iLayerType = win32con.PFD_MAIN_PLANE
        pfd.bReserved = 0
        pfd.dwLayerMask = 0
        pfd.dwVisibleMask = 0
        pfd.dwDamageMask = 0

        pixelFormat = ChoosePixelFormat(hdc, ctypes.byref(pfd))
        SetPixelFormat(hdc, pixelFormat, ctypes.byref(pfd))

        self.hrc = wglCreateContext(hdc)
        wglMakeCurrent(hdc, self.hrc)

        if version or profile:
            p = wglGetProcAddress("wglCreateContextAttribsARB".encode('utf-8'))
            if not p:
                raise Exception()
            wglCreateContextAttribsARB = ctypes.WINFUNCTYPE(ctypes.c_void_p, wintypes.HDC, ctypes.c_void_p,
                                                            ctypes.POINTER(ctypes.c_int))(p)

            attribs = []
            if version:
                major, minor = version.split('.')
                attribs.append(win32con.WGL_CONTEXT_MAJOR_VERSION_ARB)
                attribs.append(int(major))
                attribs.append(win32con.WGL_CONTEXT_MINOR_VERSION_ARB)
                attribs.append(int(minor))

            attribs.append(win32con.WGL_CONTEXT_FLAGS_ARB)
            attribs.append(0)

            if profile == 'core':
                attribs.append(win32con.WGL_CONTEXT_PROFILE_MASK_ARB)
                attribs.append(win32con.WGL_CONTEXT_CORE_PROFILE_BIT_ARB)
            elif profile == 'compatibility':
                attribs.append(win32con.WGL_CONTEXT_PROFILE_MASK_ARB)
                attribs.append(
                    win32con.WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB)

            attribs.append(0)

            array = (ctypes.c_int * len(attribs))(*attribs)
            hrc = wglCreateContextAttribsARB(
                hdc, None, array)
            wglDeleteContext(self.hrc)
            self.hrc = hrc
            wglMakeCurrent(hdc, self.hrc)

        for x in [
                'GL_VERSION',
                # 'GL_EXTENSIONS',
                'GL_SHADING_LANGUAGE_VERSION',
                'GL_RENDERER'
        ]:
            logger.info('%s: %s', x, GL.glGetString(getattr(GL, x)))

        # GL_ARB_compatibility
        profile = None
        try:
            extensions = GL.glGetString(GL.GL_EXTENSIONS).decode(  # type: ignore
                'ascii').split()
            if 'GL_ARB_compatibility' in extensions:
                profile = 'COMPATIBILITY'
        except OpenGL.error.GLerror as ex:
            profile = 'CORE'
        logger.info('OpenGL%d.%d %s', GL.glGetInteger(GL.GL_MAJOR_VERSION),
                    GL.glGetInteger(GL.GL_MINOR_VERSION), profile)

    @staticmethod
    def WndProcProxy(hwnd, message, wParam, lParam):
        # sys.stderr.write("Window::WndProcProxy\n")
        p = GetWindowLongPtr(hwnd, win32con.GWL_USERDATA)
        window = ctypes.cast(p, ctypes.py_object)
        return window.value.WndProc(hwnd, message, wParam, lParam)


class WindowFactory(object):
    def __init__(self):
        self.classes = []
        self.windows = []
        self.wndclass = self.register_class(b"MainWin")

    def __del__(self):
        self.finalize()

    def finalize(self):
        for w in self.windows:
            w.value.finalize()
        logger.debug('finalize %s', self.__class__)

    def register_class(self, className):
        """
        RegisterClassA
        """
        # Define Window Class
        wndclass = WNDCLASS()
        wndclass.style = win32con.CS_HREDRAW | win32con.CS_VREDRAW
        wndclass.lpfnWndProc = WNDPROC(WindowFactory.WndProc)
        wndclass.cbClsExtra = wndclass.cbWndExtra = 0
        wndclass.hInstance = ctypes.windll.kernel32.GetModuleHandleA(
            wintypes.HANDLE(win32con.NULL))
        wndclass.hIcon = ctypes.windll.user32.LoadIconA(
            wintypes.HANDLE(win32con.NULL), wintypes.LPCSTR(win32con.IDI_APPLICATION))
        wndclass.hCursor = ctypes.windll.user32.LoadCursorA(
            wintypes.HANDLE(win32con.NULL), wintypes.LPCSTR(win32con.IDC_ARROW))
        # wndclass.hbrBackground = ctypes.windll.gdi32.GetStockObject(
        #     ctypes.c_int(win32con.WHITE_BRUSH))
        wndclass.hbrBackground = None
        wndclass.lpszMenuName = None
        wndclass.lpszClassName = className
        # Register Window Class
        res = RegisterClass(ctypes.byref(wndclass))
        # logger.debug(res)
        if not res:
            raise ctypes.WinError()
        self.classes.append(wndclass)
        return wndclass

    def create(self, klass,
               width=win32con.CW_USEDEFAULT,
               height=win32con.CW_USEDEFAULT,
               wndclass=None,
               title=b"glglue.wgl",
               ):
        if not wndclass:
            wndclass = self.wndclass

        window = klass()
        pywindow = ctypes.py_object(window)
        hwnd = CreateWindowEx(0, wndclass.lpszClassName, title,
                              win32con.WS_OVERLAPPEDWINDOW,
                              win32con.CW_USEDEFAULT, win32con.CW_USEDEFAULT,
                              width, height, win32con.NULL, win32con.NULL,
                              wndclass.hInstance, pywindow)
        self.windows.append(pywindow)
        return window

    def loop(self, window):
        msg = MSG()
        pMsg = ctypes.pointer(msg)
        NULL = ctypes.c_int(win32con.NULL)
        # while ctypes.windll.user32.GetMessageA( pMsg, NULL, 0, 0) != 0:
        #    ctypes.windll.user32.TranslateMessage(pMsg)
        #    ctypes.windll.user32.DispatchMessageA(pMsg)
        lastCount = timeGetTime()
        while True:
            if ctypes.windll.user32.PeekMessageA(pMsg, NULL, 0, 0,
                                                 win32con.PM_NOREMOVE) != 0:
                if ctypes.windll.user32.GetMessageA(pMsg, NULL, 0, 0) == 0:
                    return msg.wParam
                ctypes.windll.user32.TranslateMessage(pMsg)
                ctypes.windll.user32.DispatchMessageA(pMsg)
            else:
                count = timeGetTime()
                d = count - lastCount
                if d > 0:
                    window.controller.onUpdate(d)
                    window.Redraw()
                    lastCount = count

    @staticmethod
    def WndProc(hwnd, message, wParam, lParam):
        # sys.stderr.write("WindowFactory::WndProc\n")
        if message == win32con.WM_CREATE:
            lpcreatestruct = ctypes.cast(lParam, ctypes.POINTER(CREATESTRUCT))
            createstruct = lpcreatestruct.contents
            window = ctypes.cast(
                createstruct.lpCreateParams, ctypes.py_object).value
            window.hwnd = hwnd
            window.SetLongPtr(hwnd, win32con.GWL_USERDATA,
                              createstruct.lpCreateParams)
            window.SetLongPtr(hwnd, win32con.GWL_WNDPROC,
                              ctypes.cast(WNDPROC(Window.WndProcProxy), ctypes.c_void_p))
            return window.WndProc(hwnd, message, wParam, lParam)

        elif message == win32con.WM_DESTROY:
            ctypes.windll.user32.PostQuitMessage(0)
            return 0

        return DefWindowProc(hwnd, message, wParam, lParam)


class LoopManager:
    def __init__(self, controller: BaseController,
                 width=win32con.CW_USEDEFAULT,
                 height=win32con.CW_USEDEFAULT,
                 wndclass=None,
                 title=b"glglue.wgl",
                 version=None,
                 profile=None,
                 ):
        self.controller = controller
        self.factory = WindowFactory()
        self.window: Window = self.factory.create(
            Window, width=width, height=height, wndclass=wndclass, title=title)
        self.window.createGLContext(16,
                                    version=version,
                                    profile=profile)
        self.window.controller = controller
        ShowWindow(self.window.hwnd, win32con.SW_SHOWNORMAL)
        self.msg = MSG()
        self.pMsg = ctypes.pointer(self.msg)
        self.NULL = ctypes.c_int(win32con.NULL)

        from OpenGL.WGL.EXT.swap_control import wglSwapIntervalEXT
        wglSwapIntervalEXT(1)

    def begin_frame(self):
        while True:
            if ctypes.windll.user32.PeekMessageA(self.pMsg, self.NULL, 0, 0,
                                                 win32con.PM_NOREMOVE) != 0:
                if ctypes.windll.user32.GetMessageA(self.pMsg, self.NULL, 0, 0) == 0:
                    return  # msg.wParam
                ctypes.windll.user32.TranslateMessage(self.pMsg)
                ctypes.windll.user32.DispatchMessageA(self.pMsg)
            break

        return timeGetTime()

    def end_frame(self):
        ps = PAINTSTRUCT()
        hdc = BeginPaint(self.window.hwnd, ctypes.byref(ps))
        if hdc:
            SwapBuffers(hdc)
        EndPaint(self.window.hwnd, ctypes.byref(ps))


def mainloop(controller, **kw):
    loop = LoopManager(controller, **kw)
    lastCount = None
    while True:
        count = loop.begin_frame()
        if not count:
            break
        d = count - lastCount if lastCount else 0
        lastCount = count
        if d > 0:
            controller.onUpdate(d)
            controller.draw()
            loop.end_frame()


if __name__ == "__main__":
    from logging import basicConfig, DEBUG
    basicConfig(format='%(levelname)s:%(name)s:%(message)s', level=DEBUG)
    import os
    if os.name != 'nt':
        logger.debug("this script is windows only: " + os.name)
        sys.exit()
    import glglue.sample
    mainloop(glglue.sample.SampleController(),
             width=600,
             height=400,
             version="3.2")



#!/usr/bin/env python
# coding: utf-8

import OpenGL.Tk
from logging import getLogger
logger = getLogger(__name__)


class Widget(OpenGL.Tk.RawOpengl):
    def __init__(self, master, engine, *args, **kw):
        #super(Widget, self).__init__(master, *args, **kw)
        OpenGL.Tk.RawOpengl.__init__(self, master, *args, **kw)
        self.engine=engine
        self.bind('<Map>', self.onDraw)
        self.bind('<Expose>', self.onDraw)
        self.bind('<Configure>', self.onResize)
        self.bind('<ButtonPress-1>', lambda e: self.engine.onLeftDown(e.x, e.y) and self.onDraw())
        self.bind('<ButtonRelease-1>', lambda e: self.engine.onLeftUp(e.x, e.y) and self.onDraw())
        self.bind('<B1-Motion>', lambda e: self.engine.onMotion(e.x, e.y) and self.onDraw())
        self.bind('<ButtonPress-2>', lambda e: self.engine.onMiddleDown(e.x, e.y) and self.onDraw())
        self.bind('<ButtonRelease-2>', lambda e: self.engine.onMiddleUp(e.x, e.y) and self.onDraw())
        self.bind('<B2-Motion>', lambda e: self.engine.onMotion(e.x, e.y) and self.onDraw())
        self.bind('<ButtonPress-3>', lambda e: self.engine.onRightDown(e.x, e.y) and self.onDraw())
        self.bind('<ButtonRelease-3>', lambda e: self.engine.onRightUp(e.x, e.y) and self.onDraw())
        self.bind('<B3-Motion>', lambda e: self.engine.onMotion(e.x, e.y) and self.onDraw())

    def onDraw(self, *dummy):
        self.tk.call(self._w, 'makecurrent')
        self.update_idletasks()
        self.engine.draw()
        self.tk.call(self._w, 'swapbuffers')

    def onResize(self, event):
        self.engine.onResize(event.width, event.height)
        self.onDraw()


if __name__=="__main__":
    import sys
    if sys.version_info[0]<3:
        import Tkinter as tkinter
    else:
        import tkinter
    import glglue.sample
    class Frame(tkinter.Frame):
        def __init__(self, width, height, master=None, **kw):
            #super(Frame, self).__init__(master, **kw)
            tkinter.Frame.__init__(self, master, **kw)
            # setup opengl widget
            self.controller=glglue.sample.SampleController()
            self.glwidget=Widget(
                    self, self.controller, width=width, height=height)
            self.glwidget.pack(fill=tkinter.BOTH, expand=True)
            # event binding(require focus)
            self.bind('<Key>', self.onKeyDown)
            self.bind('<MouseWheel>', lambda e: self.controller.onWheel(-e.delta) and self.glwidget.onDraw())

        def onKeyDown(self, event):
            key=event.keycode
            if key==27:
                # Escape
                sys.exit()
            if key==81:
                # q
                sys.exit()
            else:
                logger.debug("keycode: %d" % key)

    f = Frame(width=600, height=600)
    f.pack(fill=tkinter.BOTH, expand=True)
    f.focus_set()
    f.mainloop()




from typing import NamedTuple
from .float3 import Float3
from .quaternion import Quaternion
from .mat4 import Mat4, Float4
from .camera import Camera, FrameState
from .aabb import AABB
from .hittest import Ray


class TRS(NamedTuple):
    trnslation: Float3
    rotation: Quaternion
    scale: Float3

    def to_matrix(self) -> Mat4:
        s = Mat4.new_scale(self.scale.x, self.scale.y, self.scale.z)
        r = Mat4.new_from_quaternion(
            self.rotation.x, self.rotation.y, self.rotation.z, self.rotation.w)
        t = Mat4.new_translation(
            self.trnslation.x, self.trnslation.y, self.trnslation.z)
        return s * r * t



# coding: utf-8
'''
Win32APIでOpenGLホストするサンプル。

* Windows専用
* 追加のインストールは不要
'''
from logging import getLogger

logger = getLogger(__name__)

from logging import basicConfig, DEBUG

basicConfig(format='%(levelname)s:%(name)s:%(message)s', level=DEBUG)

from OpenGL.GL import *  # pylint: disable=W0614, W0622, W0401


class Controller:
    """
    [CLASSES] Controllerクラスは、glglueの規約に沿って以下のコールバックを実装する
    """
    def __init__(self) -> None:
        self.is_initialized = False

    def onResize(self, w: int, h: int) -> None:
        logger.debug('onResize: %d, %d', w, h)
        glViewport(0, 0, w, h)

    def onLeftDown(self, x: int, y: int) -> None:
        logger.debug('onLeftDown: %d, %d', x, y)

    def onLeftUp(self, x: int, y: int) -> None:
        logger.debug('onLeftUp: %d, %d', x, y)

    def onMiddleDown(self, x: int, y: int) -> None:
        logger.debug('onMiddleDown: %d, %d', x, y)

    def onMiddleUp(self, x: int, y: int) -> None:
        logger.debug('onMiddleUp: %d, %d', x, y)

    def onRightDown(self, x: int, y: int) -> None:
        logger.debug('onRightDown: %d, %d', x, y)

    def onRightUp(self, x: int, y: int) -> None:
        logger.debug('onRightUp: %d, %d', x, y)

    def onMotion(self, x: int, y: int) -> None:
        logger.debug('onMotion: %d, %d', x, y)

    def onWheel(self, d: int) -> None:
        logger.debug('onWheel: %d', d)

    def onKeyDown(self, keycode: int) -> None:
        logger.debug('onKeyDown: %d', keycode)

    def onUpdate(self, d: int) -> None:
        '''
        milliseconds
        '''
        #logger.debug('onUpdate: delta %d ms', d)
        pass

    def initialize(self) -> None:
        self.is_initialized = True

    def draw(self) -> None:
        if not self.is_initialized:
            self.initialize()
        glClearColor(0.0, 0.0, 1.0, 0.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glBegin(GL_TRIANGLES)
        glVertex(-1.0, -1.0)
        glVertex(1.0, -1.0)
        glVertex(0.0, 1.0)
        glEnd()

        glFlush()


if __name__ == "__main__":
    controller = Controller()
    import glglue.wgl
    loop = glglue.wgl.LoopManager(controller,
                                  width=640,
                                  height=480,
                                  title=b"sample")
    
    lastCount = 0
    while True:
        count = loop.begin_frame()
        if not count:
            break
        d = count - lastCount
        lastCount = count
        if d > 0:
            controller.onUpdate(d)
            controller.draw()
            loop.end_frame()



import logging
from OpenGL import GL
logger = logging.getLogger(__name__)


if __name__ == "__main__":
    from glglue.gl3.samplecontroller import SampleController
    controller = SampleController()

    import glglue.glfw
    loop = glglue.glfw.LoopManager(controller, title='glfw sample')

    lastCount = 0
    while True:
        count = loop.begin_frame()
        if not count:
            break
        d = count - lastCount
        lastCount = count
        if d > 0:
            controller.onUpdate(d)
            controller.draw()
            loop.end_frame()



from typing import Iterable, Dict, Optional, List
import logging
import pathlib
import ctypes
#
import glglue.glfw
from glglue.gl3.pydearcontroller import PydearController
from glglue.gl3.renderview import RenderView
from glglue.windowconfig import WindowConfig
import pydear as ImGui
from pydear.utils.dockspace import DockView
from glglue.scene.node import Node
from glglue.ctypesmath import TRS, Mat4, Camera

logger = logging.getLogger(__name__)

CONFIG_FILE = pathlib.Path("window.ini")


class SceneTree:
    def __init__(self, root: Node) -> None:
        self.root = root
        self.selected = None

    def traverse(self, node: Node):
        flag = ImGui.ImGuiTreeNodeFlags_.OpenOnArrow | ImGui.ImGuiTreeNodeFlags_.OpenOnDoubleClick | ImGui.ImGuiTreeNodeFlags_.SpanAvailWidth
        if not node.children:
            flag |= ImGui.ImGuiTreeNodeFlags_.Leaf
            flag |= ImGui.ImGuiTreeNodeFlags_.Bullet
        if node == self.selected:
            flag |= ImGui.ImGuiTreeNodeFlags_.Selected
        open = ImGui.TreeNodeEx(node.name, flag)
        if ImGui.IsItemClicked():
            # update selectable
            logger.debug(f'select: {node.name}')
            self.selected = node
        if open:
            for child in node.children:
                self.traverse(child)
            ImGui.TreePop()

    def draw(self, p_open: ctypes.Array):
        if ImGui.Begin(self.root.name, p_open):
            self.traverse(self.root)
        ImGui.End()

        # update scene
        self.root.calc_world()


class NodeProp:
    def __init__(self, get_selected, camera: Camera) -> None:
        self.get_selected = get_selected

        # ImGuizmo
        self.camera = camera
        self.matrixTranslation = (ctypes.c_float*3)()
        self.matrixRotation = (ctypes.c_float*3)()
        self.matrixScale = (ctypes.c_float*3)()
        self.mCurrentGizmoOperation = ImGui.OPERATION.ROTATE
        self.mCurrentGizmoMode = ImGui.MODE.WORLD
        self.useSnap = False

    def draw(self, p_open: ctypes.Array):
        if ImGui.Begin('selected node', p_open):
            node: Node = self.get_selected()
            if node:
                ImGui.TextUnformatted(node.name)
                self.edit_transform(node.local_transform)
                # match node.local_transform:
                #     case TRS(t, r, s):
                #         ImGui.SliderFloat3(b'T', t, -10, 10)
                #         ImGui.SliderFloat4(b'R', r, -1, 1)
                #         ImGui.SliderFloat3(b'S', s, -10, 10)
                #     case Mat4() as m:
                #         a = ctypes.addressof(m)
                #         p = ctypes.c_void_p(a)
                #         ImGui.ImGuizmo_DecomposeMatrixToComponents(
                #             p, self.matrixTranslation, self.matrixRotation, self.matrixScale)
                #         ImGui.InputFloat3(b"Tr", self.matrixTranslation)
                #         ImGui.InputFloat3(b"Rt", self.matrixRotation)
                #         ImGui.InputFloat3(b"Sc", self.matrixScale)
                #         ImGui.ImGuizmo_RecomposeMatrixFromComponents(
                #             self.matrixTranslation, self.matrixRotation, self.matrixScale, p)
        ImGui.End()

    def edit_transform(self, m: Mat4):
        ImGui.ImGuizmo_BeginFrame()
        ImGui.ImGuizmo_SetDrawlist()

        # if (ImGui.IsKeyPressed(90))
        #     mCurrentGizmoOperation = ImGui.OPERATION.TRANSLATE
        # if (ImGui.IsKeyPressed(69))
        #     mCurrentGizmoOperation = ImGui.OPERATION.ROTATE
        # if (ImGui.IsKeyPressed(82)) // r Key
        #     mCurrentGizmoOperation = ImGui.OPERATION.SCALE

        if ImGui.RadioButton("Translate", self.mCurrentGizmoOperation == ImGui.OPERATION.TRANSLATE):
            self.mCurrentGizmoOperation = ImGui.OPERATION.TRANSLATE
        ImGui.SameLine()
        if ImGui.RadioButton("Rotate", self.mCurrentGizmoOperation == ImGui.OPERATION.ROTATE):
            self.mCurrentGizmoOperation = ImGui.OPERATION.ROTATE
        ImGui.SameLine()
        if ImGui.RadioButton("Scale", self.mCurrentGizmoOperation == ImGui.OPERATION.SCALE):
            self.mCurrentGizmoOperation = ImGui.OPERATION.SCALE

        ImGui.ImGuizmo_DecomposeMatrixToComponents(
            m, self.matrixTranslation, self.matrixRotation, self.matrixScale)
        ImGui.InputFloat3("Tr", self.matrixTranslation, 3)
        ImGui.InputFloat3("Rt", self.matrixRotation, 3)
        ImGui.InputFloat3("Sc", self.matrixScale, 3)
        ImGui.ImGuizmo_RecomposeMatrixFromComponents(
            self.matrixTranslation, self.matrixRotation, self.matrixScale, m)

        if self.mCurrentGizmoOperation != ImGui.OPERATION.SCALE:
            if ImGui.RadioButton("Local", self.mCurrentGizmoMode == ImGui.MODE.LOCAL):
                self.mCurrentGizmoMode = ImGui.MODE.LOCAL
            ImGui.SameLine()
            if ImGui.RadioButton("World", self.mCurrentGizmoMode == ImGui.MODE.WORLD):
                self.mCurrentGizmoMode = ImGui.MODE.WORLD

        # if (ImGui.IsKeyPressed(83))
        #     useSnap = !useSnap
        # ImGui.Checkbox("", &useSnap)
        # ImGui.SameLine()
        # vec_t snap
        # switch (mCurrentGizmoOperation)
        # {
        # case ImGui.OPERATION.TRANSLATE:
        #     snap = config.mSnapTranslation
        #     ImGui.InputFloat3("Snap", &snap.x)
        #     break
        # case ImGui.OPERATION.ROTATE:
        #     snap = config.mSnapRotation
        #     ImGui.InputFloat("Angle Snap", &snap.x)
        #     break
        # case ImGui.OPERATION.SCALE:
        #     snap = config.mSnapScale
        #     ImGui.InputFloat("Scale Snap", &snap.x)
        #     break
        # }

        io = ImGui.GetIO()
        # x, y = ImGui.GetWindowPos()
        # w, h = ImGui.GetWindowSize()
        # ImGui.ImGuizmo_SetRect(x, y, w, h)
        ImGui.ImGuizmo_SetRect(0, 0, io.DisplaySize.x, io.DisplaySize.y)
        ImGui.ImGuizmo_Manipulate(self.camera.view.matrix, self.camera.projection.matrix, self.mCurrentGizmoOperation, self.mCurrentGizmoMode, m, None, None)#self.useSnap ? &snap.x : NULL)


def cube():
    view = RenderView()
    return [DockView(
        'cube', (ctypes.c_bool * 1)(True), view.draw)]


def teapot():
    view = RenderView()
    from glglue.scene.teapot import create_teapot
    view.scene.drawables = [create_teapot()]
    return [DockView(
        'teapot', (ctypes.c_bool * 1)(True), view.draw)]


def skin():
    view = RenderView()
    from glglue.scene.skin_sample import create_skin

    root = create_skin()
    view.scene.drawables = [root]

    tree = SceneTree(root)
    # prop = NodeProp(lambda: tree.selected, view.camera)

    return [
        DockView('skin', (ctypes.c_bool * 1)(True), view.draw),
        DockView('skin_herarchy', (ctypes.c_bool * 1)(True), tree.draw),
        # DockView('skin_node_prop', (ctypes.c_bool * 1)(True), prop.draw),
    ]


SCENES = {
    'cube': cube,
    'teapot': teapot,
    'skin': skin,
}


class ImguiDocks:
    def __init__(self) -> None:
        self.demo = DockView(
            'demo', (ctypes.c_bool * 1)(True), ImGui.ShowDemoWindow)
        self.metrics = DockView(
            'metrics', (ctypes.c_bool * 1)(True), ImGui.ShowMetricsWindow)
        self.selected = 'skin'
        self.scenes: Dict[str, List[DockView]] = {
            k: [] for k, v in SCENES.items()}

        def show_selector(p_open: ctypes.Array):
            # open new window context
            if ImGui.Begin("SceneSelector", p_open):
                for k, v in self.scenes.items():
                    if ImGui.Selectable(k, k == self.selected):
                        self.selected = k
                # draw text label inside of current window
                if ImGui.Button("Debug"):
                    logger.debug("debug message")
            # close current window context
            ImGui.End()
        self.hello = DockView(
            'hello', (ctypes.c_bool * 1)(True), show_selector)

        # logger
        from pydear.utils.loghandler import ImGuiLogHandler
        log_handle = ImGuiLogHandler()
        log_handle.register_root()
        self.logger = DockView('log', (ctypes.c_bool * 1)
                               (True), log_handle.draw)

    def get_or_create(self, key: str) -> Iterable[DockView]:
        value = self.scenes.get(key)
        if value:
            return value

        # create RenderView
        logger.info(f"create: {key}")
        value = SCENES[key]()
        self.scenes[key] = value
        return value

    def __iter__(self) -> Iterable[DockView]:
        yield self.demo
        yield self.metrics
        yield self.hello
        for dock in self.get_or_create(self.selected):
            yield dock
        yield self.logger


class SampleController(PydearController):
    def imgui_create_docks(self):
        return ImguiDocks()


if __name__ == "__main__":
    logging.basicConfig(
        format='%(levelname)s:%(name)s:%(message)s', level=logging.DEBUG)

    # ImGui
    controller = SampleController()

    # glfw
    loop = glglue.glfw.LoopManager(
        controller,
        config=WindowConfig.load_json_from_path(CONFIG_FILE),
        title="pydear")

    # main loop
    lastCount = 0
    while True:
        count = loop.begin_frame()
        if not count:
            break
        d = count - lastCount
        lastCount = count
        if d > 0:
            controller.onUpdate(d)
            controller.draw()
            loop.end_frame()

    # save window config
    config = loop.get_config()
    config.save_json_to_path(CONFIG_FILE)



from typing import List, Any
from .vertices import VectorView


class Skin:
    def __init__(self, joints: List[Any], inverse_bind_matrices: VectorView) -> None:
        from .node import Node
        self.joints: List[Node] = joints
        self.bind_matrices = inverse_bind_matrices



import sys
sys.path.append('.')
sys.path.append('..')

import os
import pathlib

if 'PYSDL2_DLL_PATH' not in os.environ:
    exe = pathlib.Path(sys.executable)
    sdl2dll = exe.parent / 'sdl2.dll'
    if sdl2dll.exists():
        os.environ['PYSDL2_DLL_PATH'] = str(exe.parent)

from sdl2 import *
import ctypes


class LoopManager:
    def __init__(self, controller, **kw):
        self.controller = controller
        title = kw.get('title', b'glglue.pysdl2')
        width = kw.get('width', 600)
        height = kw.get('height', 400)
        SDL_Init(SDL_INIT_VIDEO)
        SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1)
        self.window = SDL_CreateWindow(title, SDL_WINDOWPOS_UNDEFINED,
                                       SDL_WINDOWPOS_UNDEFINED, width, height,
                                       SDL_WINDOW_OPENGL)
        SDL_GL_CreateContext(self.window)
        self.controller.onResize(width, height)
        self.event = SDL_Event()

    def begin_frame(self):
        while SDL_PollEvent(ctypes.byref(self.event)) != 0:
            if self.event.type == SDL_QUIT:
                return
            elif self.event.type == SDL_KEYDOWN:
                self.controller.onKeyDown(self.event.key.keysym.sym)
            elif self.event.type == SDL_MOUSEBUTTONDOWN:
                if self.event.button.button == SDL_BUTTON_LEFT:
                    self.controller.onLeftDown(self.event.button.x,
                                               self.event.button.y)
                elif self.event.button.button == SDL_BUTTON_MIDDLE:
                    self.controller.onMiddleDown(self.event.button.x,
                                                 self.event.button.y)
                elif self.event.button.button == SDL_BUTTON_RIGHT:
                    self.controller.onRightDown(self.event.button.x,
                                                self.event.button.y)
            elif self.event.type == SDL_MOUSEBUTTONUP:
                if self.event.button.button == SDL_BUTTON_LEFT:
                    self.controller.onLeftUp(self.event.button.x,
                                             self.event.button.y)
                elif self.event.button.button == SDL_BUTTON_MIDDLE:
                    self.controller.onMiddleUp(self.event.button.x,
                                               self.event.button.y)
                elif self.event.button.button == SDL_BUTTON_RIGHT:
                    self.controller.onRightUp(self.event.button.x,
                                              self.event.button.y)
            elif self.event.type == SDL_MOUSEMOTION:
                self.controller.onMotion(self.event.motion.x,
                                         self.event.motion.y)
            elif self.event.type == SDL_MOUSEWHEEL:
                self.controller.onWheel(-self.event.wheel.y)

        return SDL_GetTicks()

    def end_frame(self):
        SDL_GL_SwapWindow(self.window)


if __name__ == "__main__":
    import glglue.sample
    controller = glglue.sample.SampleController()
    loop = LoopManager(controller)
    lastTime = 0
    while True:
        time = loop.begin_frame()
        if not time:
            break
        if lastTime == 0:
            lastTime = time
            continue
        d = time - lastTime
        lastTime = time
        if d:
            controller.onUpdate(d)
            controller.draw()
            loop.end_frame()



import ctypes
from OpenGL import GL
from glglue.ctypesmath import *
import glglue.scene.vertices
import glglue.gl3.shader
import glglue.gl3.vbo
import logging
logger = logging.getLogger(__name__)


class Vertex(ctypes.Structure):
    _fields_ = [
        ('position', Float3),
        ('color', Float4)
    ]


VS = '''
#version 330
in vec3 aPosition;
in vec4 aColor;
out vec4 vColor;
uniform mediump mat4 vp;

void main() {
  gl_Position = vec4(aPosition, 1) * vp;
  vColor = aColor;
}
'''

FS = '''
#version 330
in vec4 vColor;
out vec4 fColor;
void main() { fColor = vColor; }
'''


class Gizmo:
    def __init__(self) -> None:
        # state
        self.state = FrameState(Float4(0, 0, 1, 1), 0, 0, False,
                                False, False, Mat4.new_identity(), Mat4.new_identity(),
                                Ray(Float3(0, 0, 0), Float3(0, 0, 1)),
                                Float4(1, 1, 1, 1))
        self.matrix = Mat4.new_identity()
        self.color = Float4(1, 1, 1, 1)
        # event
        self.click_left = False
        self.click_middle = False
        self.click_right = False
        # lines
        self.lines = (Vertex * 65535)()
        self.line_count = 0
        self.line_shader = None
        self.line_drawable = None
        # triangles
        self.triangles = (Vertex * 65535)()
        self.triangle_count = 0
        self.triangle_shader = None
        self.triangle_drawable = None

        # hover selectable
        self.hover = None
        self.hover_last = None

    def begin(self, state: FrameState):
        # clear
        self.line_count = 0
        self.triangle_count = 0
        self.matrix = Mat4.new_identity()
        self.color = Float4(1, 1, 1, 1)
        # update
        self.click_left = self.state.mouse_left_down and not state.mouse_left_down
        self.click_right = self.state.mouse_right_down and not state.mouse_right_down
        self.click_middle = self.state.mouse_middle_down and not state.mouse_middle_down
        self.state = state

        self.hover_last = self.hover
        self.hover = None

    def end(self):
        # material
        if not self.line_shader:
            shader_source = glglue.gl3.shader.ShaderSource(
                VS, (), FS, ())
            self.line_shader = glglue.gl3.shader.create_from(shader_source)
        self.line_shader.use()
        vp = self.state.camera_view * self.state.camera_projection
        self.line_shader.set_uniform('vp', vp)

        self._draw_triangles()
        self._draw_lines()

    def _draw_lines(self):
        if self.line_drawable:
            self.line_drawable.vbo_list[0].update(memoryview(self.lines))
        else:
            typed = glglue.scene.vertices.VectorView(
                memoryview(self.lines), ctypes.c_float, 7)
            self.line_drawable = glglue.gl3.vbo.create(
                glglue.gl3.vbo.Interleaved(typed, [0, 12]), is_dynamic=True)
        self.line_drawable.draw(GL.GL_LINES, 0, self.line_count)

    def _draw_triangles(self):
        if self.triangle_drawable:
            self.triangle_drawable.vbo_list[0].update(
                memoryview(self.triangles))
        else:
            typed = glglue.scene.vertices.VectorView(
                memoryview(self.triangles), ctypes.c_float, 7)
            self.triangle_drawable = glglue.gl3.vbo.create(
                glglue.gl3.vbo.Interleaved(typed, [0, 12]), is_dynamic=True)

        GL.glDisable(GL.GL_DEPTH_TEST)
        GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
        GL.glEnable(GL.GL_BLEND)
        self.triangle_drawable.draw(GL.GL_TRIANGLES, 0, self.triangle_count)
        GL.glDisable(GL.GL_BLEND)
        GL.glEnable(GL.GL_DEPTH_TEST)

    def line(self, p0: Float3, p1: Float3):
        p0 = self.matrix.apply(*p0)
        self.lines[self.line_count] = Vertex(p0, self.color)
        self.line_count += 1

        p1 = self.matrix.apply(*p1)
        self.lines[self.line_count] = Vertex(p1, self.color)
        self.line_count += 1

    def triangle(self, p0: Float3, p1: Float3, p2: Float3, *, intersect=False):
        p0 = self.matrix.apply(*p0)
        p1 = self.matrix.apply(*p1)
        p2 = self.matrix.apply(*p2)

        self.triangles[self.triangle_count] = Vertex(p0, self.color)
        self.triangle_count += 1

        self.triangles[self.triangle_count] = Vertex(p1, self.color)
        self.triangle_count += 1

        self.triangles[self.triangle_count] = Vertex(p2, self.color)
        self.triangle_count += 1

        if intersect:
            return self.state.ray.intersect(p0, p1, p2)

    def quad(self, p0: Float3, p1: Float3, p2: Float3, p3: Float3):
        self.triangle(p0, p1, p2)
        self.triangle(p2, p3, p0)

    def axis(self, size: float):
        origin = Float3(0, 0, 0)
        # X
        self.color = Float4(1, 0, 0, 1)
        self.line(origin, Float3(size, 0, 0))
        self.color = Float4(0.5, 0, 0, 1)
        self.line(origin, Float3(-size, 0, 0))
        # Y
        self.color = Float4(0, 1, 0, 1)
        self.line(origin, Float3(0, size, 0))
        self.color = Float4(0, 0.5, 0, 1)
        self.line(origin, Float3(0, -size, 0))
        # Z
        self.color = Float4(0, 0, 1, 1)
        self.line(origin, Float3(0, 0, size))
        self.color = Float4(0, 0, 0.5, 1)
        self.line(origin, Float3(0, 0, -size))

    def aabb(self, aabb: AABB):
        self.color = Float4(1, 1, 1, 1)
        match aabb:
            case AABB(Float3(nx, ny, nz), Float3(px, py, pz)):
                t0 = Float3(nx, py, nz)
                t1 = Float3(px, py, nz)
                t2 = Float3(px, py, pz)
                t3 = Float3(nx, py, pz)
                b0 = Float3(nx, ny, nz)
                b1 = Float3(px, ny, nz)
                b2 = Float3(px, ny, pz)
                b3 = Float3(nx, ny, pz)
                # top
                self.line(t0, t1)
                self.line(t1, t2)
                self.line(t2, t3)
                self.line(t3, t0)
                # bottom
                self.line(b0, b1)
                self.line(b1, b2)
                self.line(b2, b3)
                self.line(b3, b0)
                # side
                self.line(t0, b0)
                self.line(t1, b1)
                self.line(t2, b2)
                self.line(t3, b3)

    def bone(self, key, length: float, is_selected: bool = False) -> bool:
        '''
        return True if mouse clicked
        '''
        s = length * 0.1
        # head-tail
        #      0, -1(p1)
        # (p2)  |
        # -1, 0 |
        #     --+--->
        #       |    1, 0(p0)
        #       v
        #      0, +1(p3)
        self.color = Float4(1, 0.0, 1, 1)
        h = Float3(0, 0, 0)
        t = Float3(0, length, 0)
        # self.line(h, t, bone.world_matrix)
        p0 = Float3(s, s, 0)
        p1 = Float3(0, s, -s)
        p2 = Float3(-s, s, 0)
        p3 = Float3(0, s, s)

        self.line(p0, p1)
        self.line(p1, p2)
        self.line(p2, p3)
        self.line(p3, p0)

        # self.line(p2, p0, bone.world_matrix)
        self.color = Float4(1, 0, 0, 1)
        self.line(h, p0)
        self.line(p0, t)
        self.color = Float4(0.1, 0, 0, 1)
        if is_selected:
            self.color = Float4(0.1, 1, 0, 1)
        self.line(h, p2)
        self.line(p2, t)

        # self.line(p1, p3, bone.world_matrix)
        self.color = Float4(0, 0, 1, 1)
        self.line(h, p3)
        self.line(p3, t)
        self.color = Float4(0, 0, 0.1, 1)
        if is_selected:
            self.color = Float4(0, 1, 0.1, 1)
        self.line(h, p1)
        self.line(p1, t)

        # triangles
        clicked = False
        self.color = Float4(0.5, 0.5, 0.5, 0.2)
        if is_selected:
            self.color = Float4(0.7, 0.7, 0, 0.7)
        elif self.hover_last == key:
            self.color = Float4(0, 0.7, 0, 0.7)
            if self.click_left:
                clicked = True

        triangles = (
            (p0, h, p1),
            (p1, h, p2),
            (p2, h, p3),
            (p3, h, p0),
            (p0, t, p1),
            (p1, t, p2),
            (p2, t, p3),
            (p3, t, p0),
        )

        any_hit = False
        for t in triangles:
            hit = self.triangle(*t, intersect=(not any_hit))
            if hit:
                self.hover = key
                any_hit = True

        return clicked



import unittest


class TestPy(unittest.TestCase):

    def test_false(self):
        # 🤔
        self.assertFalse(0)
        self.assertFalse(0.0)
        self.assertFalse([])
        self.assertFalse(())
        self.assertFalse({})


if __name__ == '__main__':
    unittest.main()



from typing import Iterable
import ctypes
import pkgutil
import struct
from OpenGL import GL
from .mesh import Mesh, Material
from .vertices import Interleaved, VectorView


class Float3(ctypes.Structure):
    _fields_ = [
        ('x', ctypes.c_float),
        ('y', ctypes.c_float),
        ('z', ctypes.c_float),
    ]

    def __iter__(self) -> Iterable[float]:
        yield self.x
        yield self.y
        yield self.z


class StlTriangle(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ('normal', Float3),
        ('position0', Float3),
        ('position1', Float3),
        ('position2', Float3),
        ('attributes', ctypes.c_ushort),
    ]


assert ctypes.sizeof(StlTriangle) == 50


class Vertex(ctypes.Structure):
    _fields_ = [
        ('position', Float3),
        ('normal', Float3),
    ]


class BinaryReader:
    def __init__(self, data: bytes) -> None:
        self.data = data
        self.pos = 0

    def read(self, size: int) -> bytes:
        if self.pos + size > len(self.data):
            raise RuntimeError()
        data = self.data[self.pos:self.pos+size]
        self.pos += size
        return data

    def read_uint32(self) -> int:
        data = self.read(4)
        return struct.unpack('I', data)[0]

    def read_float32(self) -> float:
        data = self.read(4)
        return struct.unpack('f', data)[0]


class Builder:
    def __init__(self, triangle_count: int):
        self.vertices = (Vertex * (triangle_count * 3))()

    def push_triangle(self, i: int, t: StlTriangle, scale: float):
        p0x, p0y, p0z = t.position0
        p1x, p1y, p1z = t.position1
        p2x, p2y, p2z = t.position2
        nx, ny, nz = t.normal
        v0 = self.vertices[i]
        v0.position = Float3(p0x*scale, p0z*scale, -p0y*scale)
        v0.normal = Float3(nx, nz, -ny)
        v1 = self.vertices[i+1]
        v1.position = Float3(p1x*scale, p1z*scale, -p1y*scale)
        v1.normal = Float3(nx, nz, -ny)
        v2 = self.vertices[i+2]
        v2.position = Float3(p2x*scale, p2z*scale, -p2y*scale)
        v2.normal = Float3(nx, nz, -ny)

    def build(self):
        mesh = Mesh('teapot', Interleaved(
            VectorView(memoryview(self.vertices), ctypes.c_float, 6), [0, 12]))

        vs = pkgutil.get_data('glglue', 'assets/teapot.vs')
        if not vs:
            raise Exception()
        fs = pkgutil.get_data('glglue', 'assets/teapot.fs')
        if not fs:
            raise Exception()
        material = Material('teapot', vs.decode('utf-8'), fs.decode('utf-8'))

        mesh.add_submesh(material, [], GL.GL_TRIANGLES)

        return mesh


def create_teapot() -> Mesh:

    # https://en.wikipedia.org/wiki/Utah_teapot#/media/File:Utah_teapot_(solid).stl
    data = pkgutil.get_data('glglue', 'assets/Utah_teapot_(solid).stl')
    if not data:
        raise RuntimeError()

    # https://en.wikipedia.org/wiki/STL_%28file_format%29
    r = BinaryReader(data)

    # UINT8[80]    – Header                 -     80 bytes
    # UINT32       – Number of triangles    -      4 bytes
    header = r.read(80)
    triangle_count = r.read_uint32()

    triangles_type = StlTriangle * triangle_count
    data = r.read(ctypes.sizeof(triangles_type))
    triangles = triangles_type.from_buffer_copy(data)

    b = Builder(triangle_count)
    for i, t in enumerate(triangles):
        b.push_triangle(i * 3, t, 0.1)

    return b.build()



from glglue.gl3.imguicontroller import ImGuiController
import logging
logger = logging.getLogger(__name__)
logging.basicConfig(
    format='%(levelname)s:%(name)s:%(message)s', level=logging.DEBUG)


class CustomGUI(ImGuiController):
    def on_imgui(self):
        import imgui

        # open new window context
        imgui.begin("CustomGUI", True)
        # draw text label inside of current window
        imgui.text("imgui !")
        # close current window context
        imgui.end()


if __name__ == "__main__":
    controller = CustomGUI()

    import glglue.wgl
    loop = glglue.wgl.LoopManager(controller,
                                  width=640,
                                  height=480,
                                  title=b"imgui")

    lastCount = 0
    while True:
        count = loop.begin_frame()
        if not count:
            break
        d = count - lastCount
        lastCount = count
        if d > 0:
            controller.onUpdate(d)
            controller.draw()
            loop.end_frame()



def get_desktop_scaling_factor():
    '''
    for high DPI desktop
    '''
    import platform
    if platform.platform().startswith('Windows'):
        from ctypes import windll
        desktop = windll.user32.GetDC(0)
        LogicalScreenHeight = windll.gdi32.GetDeviceCaps(
            desktop, 10)  # VERTRES
        PhysicalScreenHeight = windll.gdi32.GetDeviceCaps(
            desktop, 117)  # DESKTOPVERTRES
        return PhysicalScreenHeight / LogicalScreenHeight

    return 1
